"""Models for LK3."""
import attr

from .validators import integer_str


@attr.s(auto_attribs=True, frozen=True)
class Info:
    """Object holding information from LK3."""
    mac_address: str
    sw_version: str
    model: str
    name: str
    brand: str

    @staticmethod
    def from_dict(data: dict):
        """Return Info object from LK3 API response."""
        return Info(
            mac_address=data.get('mac', ''),
            sw_version=data.get('sw', 'Unknown'),
            model=data.get('hw', 'Unknown'),
            name='LK3',
            brand='Tinycontrol',
        )


@attr.s(auto_attribs=True, frozen=True)
class State:
    """Object holding information from LK3 sensors."""
    out0: bool
    out1: bool
    out2: bool
    out3: bool
    out4: bool
    out5: bool
    pwm0: bool
    pwm1: bool
    pwm2: bool
    pwm3: bool
    event1: bool
    event2: bool
    event3: bool
    event4: bool
    event5: bool
    event6: bool
    event7: bool
    event8: bool
    vin: float
    tem: float
    ds1: float
    ds2: float
    ds3: float
    ds4: float
    ds5: float
    ds6: float
    ds7: float
    ds8: float
    t1: float
    h1: float
    p1: float
    pm1: float
    pm2: float
    pm4: float
    pm10: float
    co2: int
    diff1: float
    diff2: float
    diff3: float
    diff4: float
    diff5: float
    diff6: float
    inpa1: float
    inpa2: float
    inpa3: float
    inpa4: float
    inpa5: float
    inpa6: float
    inpd1: bool
    inpd2: bool
    inpd3: bool
    inpd4: bool
    power1: float
    power2: float
    power3: float
    power4: float
    power5: float
    power6: float
    energy1: float
    energy2: float
    energy3: float
    energy4: float
    energy5: float
    energy6: float
    duralux0: float
    duralux1: float
    duralux2: float
    duralux3: float
    duralux4: float
    duralux5: float
    duralux6: float
    duralux7: float
    duralux8: float
    m1: float
    m2: float
    m3: float
    m4: float
    m5: float
    m6: float
    m7: float
    m8: float
    m9: float
    m10: float
    m11: float
    m12: float
    m13: float
    m14: float
    m15: float
    m16: float
    m17: float
    m18: float
    m19: float
    m20: float
    m21: float
    m22: float
    m23: float
    m24: float
    m25: float
    m26: float
    m27: float
    m28: float
    m29: float
    m30: float

    def __getitem__(self, key):
        """Make object subscriptable (self[key])."""
        return self.__dict__[key]

    def items(self):
        return self.__dict__.items()

    @staticmethod
    def decimal_to_bool_list(data, length):
        """Convert decimal string to list of bool values.

        Eg. decimal_to_bool_list('3', 4) -> [True, True, False, False]
        """
        return [
            bool(int(item))
            for item in reversed(('{:0' + str(length) + 'b}').format(int(data)))
        ]

    @staticmethod
    def parse_diff(value, source, diffsel, depth=3):
        """Parse value of diff according to source sensor.

        Valid till SW 1.47a
        """
        if depth == 0:
            return 0
        if source < 6 or source == 24:
            return value/100
        elif source < 14 or source == 22 or source == 23:
            return value/10
        elif source >= 25 and source <= 27:
            return State.parse_diff(value, diffsel[source-25], diffsel, depth-1)
        else:
            return value/1000

    @staticmethod
    def parse_custom_readings(data):
        """Parse readings mappings (Modbus/1 Wire) and HW3.0 SDMs."""
        sw_version = data.get('sw')
        hw_version = data.get('hw')
        if hw_version is not None and sw_version is not None:
            if hw_version == '3.0' or (hw_version != '3.0' and sw_version <= '1.31'):
                # Dict method for sdm1-sdm14, rest is unknown
                return [int(data.get(f'sdm{i}', 0))/100 for i in range(1, 31)]
            elif hw_version != '3.0' and sw_version < '1.33':
                # List method for sdm1-sdm29 for hw 3.5+ sw 1.32+
                modbus_sensor = int(data.get('modbusSensor', 0))
                sdm_list = [int(item) for item in data.get('sdm')]
                if modbus_sensor == 5:
                    return [
                        *sdm_list[0:6], # 0-5
                        sdm_list[6]/10,
                        sdm_list[7]/100,
                        sdm_list[8]/10,
                        sdm_list[9]/100,
                        sdm_list[10]/100,
                        sdm_list[11]/100,
                        sdm_list[12]/100,
                        sdm_list[13]/100,
                        sdm_list[14]/100,
                        sdm_list[15]/10,
                        sdm_list[16]/100,
                        sdm_list[17]/10,
                        sdm_list[18]/100,
                        sdm_list[19]/10,
                        sdm_list[20]/100,
                        *sdm_list[21:25], # 21-24
                        sdm_list[25]/100,
                        *sdm_list[26:29], # 26-28
                        0,
                    ]
                elif modbus_sensor == 4:
                    return [
                        *[item/100 for item in sdm_list[0:9]], # 0-8
                        *sdm_list[9:13], # 9-12
                        *[item/100 for item in sdm_list[13:29]], # 13-28
                        0,
                    ]
                else:
                    return [item/100 for item in sdm_list] + [0]
            elif hw_version != '3.0' and sw_version < '1.50':
                # hw 3.5+ sw 1.33+ (3 modbus slots)
                modbus_mapping = [(int(item[0]), int(item[1:]))
                                  for item in data.get('modbusMapping').split('*')]
                modbus_values = data.get('modbusReadings')
                mapped_values = []
                for item in modbus_mapping:
                    if item[0] == 0:
                        mapped_values.append(0)
                    else:
                        # Slots are 1-3, so `-1`
                        mapped_values.append(float(modbus_values[item[0]-1][item[1]]))
                return mapped_values
            else:
                return [float(item) for item in data.get('customReadings')]
        return [0]*30

    @staticmethod
    def parse_diffs(data):
        """Parse diffs from data."""
        diffs = [0]*6
        sw_version = data.get('sw')
        hw_version = data.get('hw')
        if hw_version is not None and sw_version is not None:
            if hw_version != '3.0' and sw_version >= '1.49':
                # Version with 6 diffs
                diffs = [
                    int(data.get(f'diff{index+1}', 0))/1000
                    for index in range(6)
                ]
            else:
                # Version with 3 diffs
                diffsel = list(map(int, data.get('diffsel', '').split('*')))
                # Guard from errors diff doesnt refer itself
                if diffsel[0] != 25 and diffsel[1] != 26 and diffsel[2] != 27:
                    diffs = [
                        State.parse_diff(int(data.get(f'diff{index+1}', 0)), diffsel[index], diffsel)
                        for index in range(3)
                    ] + [0]*3
        return diffs

    @staticmethod
    def negation(value, condition):
        """Negate boolean value if conditon is True."""
        if condition:
            return not value
        return value

    @staticmethod
    def parse_digital_inputs(data):
        """Parse digital inputs, applying negation if needed."""
        sw_version = data.get('sw')
        hw_version = data.get('hw')
        inpd = State.decimal_to_bool_list(data.get('ind', 0), 4)
        if hw_version != '3.0' and sw_version >= '1.49':
            return inpd
        else:
            digital_inputs_negation = State.decimal_to_bool_list(data.get('inpdnn', 0), 4)
            print(data.get('inpdnn', 0), inpd, digital_inputs_negation)
            return [
                State.negation(value, negation)
                for value, negation in zip(inpd, digital_inputs_negation)
            ]

    @staticmethod
    def parse_event_variables(data):
        """Return state of event variables."""
        events = data.get('eventVariables')
        if events is None:
            # Fallback for older version (works properly for 3.5+ 1.22b-1.39)
            events = data.get('eveOutStates', '0*0*0*0').split('*')[2]
            if not integer_str(events):
                events = 0
        return State.decimal_to_bool_list(events, 8)

    @staticmethod
    def from_dict(data):
        """Return State object from LK3 API response."""
        pwm = State.decimal_to_bool_list(data.get('pwm', 0), 4)
        digital_inputs = State.parse_digital_inputs(data)
        diffs = State.parse_diffs(data)
        custom_readings = State.parse_custom_readings(data)
        outputs_negation = bool(int(data.get('outnn', '0')))
        events = State.parse_event_variables(data)
        return State(
            out0=State.negation(bool(int(data.get('out0', 0))), outputs_negation),
            out1=State.negation(bool(int(data.get('out1', 0))), outputs_negation),
            out2=State.negation(bool(int(data.get('out2', 0))), outputs_negation),
            out3=State.negation(bool(int(data.get('out3', 0))), outputs_negation),
            out4=State.negation(bool(int(data.get('out4', 0))), outputs_negation),
            out5=State.negation(bool(int(data.get('out5', 0))), outputs_negation),
            pwm0=pwm[0],
            pwm1=pwm[1],
            pwm2=pwm[2],
            pwm3=pwm[3],
            event1=events[0],
            event2=events[1],
            event3=events[2],
            event4=events[3],
            event5=events[4],
            event6=events[5],
            event7=events[6],
            event8=events[7],
            vin=int(data.get('vin', 0))/100,
            tem=int(data.get('tem', 0))/100,
            ds1=int(data.get('ds1', 0))/10,
            ds2=int(data.get('ds2', 0))/10,
            ds3=int(data.get('ds3', 0))/10,
            ds4=int(data.get('ds4', 0))/10,
            ds5=int(data.get('ds5', 0))/10,
            ds6=int(data.get('ds6', 0))/10,
            ds7=int(data.get('ds7', 0))/10,
            ds8=int(data.get('ds8', 0))/10,
            t1=int(data.get('dthTemp', 0))/10,
            h1=int(data.get('dthHum', 0))/10,
            p1=int(data.get('bm280p', 0))/100,
            pm1=int(data.get('pm1', 0))/10,
            pm2=int(data.get('pm2', 0))/10,
            pm4=int(data.get('pm4', 0))/10,
            pm10=int(data.get('pm10', 0))/10,
            co2=int(data.get('co2', 0)),
            diff1=diffs[0],
            diff2=diffs[1],
            diff3=diffs[2],
            diff4=diffs[3],
            diff5=diffs[4],
            diff6=diffs[5],
            inpa1=int(data.get('inpp1', 0))/100,
            inpa2=int(data.get('inpp2', 0))/100,
            inpa3=int(data.get('inpp3', 0))/100,
            inpa4=int(data.get('inpp4', 0))/100,
            inpa5=int(data.get('inpp5', 0))/100,
            inpa6=int(data.get('inpp6', 0))/100,
            inpd1=digital_inputs[0],
            inpd2=digital_inputs[1],
            inpd3=digital_inputs[2],
            inpd4=digital_inputs[3],
            power1=int(data.get('power1', 0))/1000,
            power2=int(data.get('power2', 0))/1000,
            power3=int(data.get('power3', 0))/1000,
            power4=int(data.get('power4', 0))/1000,
            power5=int(data.get('power5', 0))/1000,
            power6=int(data.get('power6', 0))/1000,
            energy1=int(data.get('energy1', 0))/1000,
            energy2=int(data.get('energy2', 0))/1000,
            energy3=int(data.get('energy3', 0))/1000,
            energy4=int(data.get('energy4', 0))/1000,
            energy5=int(data.get('energy5', 0))/1000,
            energy6=int(data.get('energy6', 0))/1000,
            duralux0=int(data.get('d0', 0))/10,
            duralux1=int(data.get('d1', 0))/10,
            duralux2=int(data.get('d2', 0))/10,
            duralux3=int(data.get('d3', 0))/10,
            duralux4=int(data.get('d4', 0))/100,
            duralux5=int(data.get('d5', 0)),
            duralux6=int(data.get('d6', 0)),
            duralux7=int(data.get('d7', 0))/10,
            duralux8=int(data.get('d8', 0)),
            **{f'm{index+1}': value for index, value in enumerate(custom_readings)}, # m1-m30
        )


@attr.s(auto_attribs=True, frozen=True)
class Device:
    """Object holding all information of LK3."""
    info: Info
    state: State

    @staticmethod
    def from_dict(data):
        """Return Device object from LK3 API response."""
        return Device(
            info=Info.from_dict(data),
            state=State.from_dict(data),
        )
