"""Support for LK3 switches."""
import logging
from typing import Any, Callable, List

from homeassistant.components.switch import SwitchEntity
from homeassistant.config_entries import ConfigEntry
from homeassistant.helpers.entity import Entity
from homeassistant.helpers.typing import HomeAssistantType

from . import LK3DeviceEntity
from .lk3api import LK3Reader, LK3Error
from .const import (
    DATA_LK3_CLIENT,
    DOMAIN,
)

_LOGGER = logging.getLogger(__name__)

PARALLEL_UPDATES = 1

SWITCH_LIST = [
    {'name': 'OUT0', 'key': 'out0'},
    {'name': 'OUT1', 'key': 'out1'},
    {'name': 'OUT2', 'key': 'out2'},
    {'name': 'OUT3', 'key': 'out3'},
    {'name': 'OUT4', 'key': 'out4'},
    {'name': 'OUT5', 'key': 'out5'},
    {'name': 'PWM0', 'key': 'pwm0'},
    {'name': 'PWM1', 'key': 'pwm1'},
    {'name': 'PWM2', 'key': 'pwm2'},
    {'name': 'PWM3', 'key': 'pwm3'},
    {'name': 'EVENT1', 'key': 'event1'},
    {'name': 'EVENT2', 'key': 'event2'},
    {'name': 'EVENT3', 'key': 'event3'},
    {'name': 'EVENT4', 'key': 'event4'},
    {'name': 'EVENT5', 'key': 'event5'},
    {'name': 'EVENT6', 'key': 'event6'},
    {'name': 'EVENT7', 'key': 'event7'},
    {'name': 'EVENT8', 'key': 'event8'},
]


async def async_setup_entry(
        hass: HomeAssistantType,
        entry: ConfigEntry,
        async_add_entities: Callable[[List[Entity], bool], None],
) -> None:
    """Set up LK3 switch based on a config entry."""
    reader: LK3Reader = hass.data[DOMAIN][entry.entry_id][DATA_LK3_CLIENT]

    switches = [
        LK3Switch(
            entry.entry_id,
            reader,
            item['name'],
            None,
            item['key'],
        ) for item in SWITCH_LIST
    ]
    async_add_entities(switches, True)


class LK3Switch(LK3DeviceEntity, SwitchEntity):
    """Defines a LK3 switch."""

    def __init__(
            self,
            entry_id: str,
            reader: LK3Reader,
            name: str,
            icon: str,
            key: str,
    ) -> None:
        """Initialize LK3 switch."""
        self._key = key
        self._state = False
        super().__init__(entry_id, reader, name, icon)

    @property
    def unique_id(self) -> str:
        """Return the unique ID for this sensor."""
        return f"{self.reader.device.info.mac_address}_{self._key}"

    @property
    def is_on(self) -> bool:
        """Return the state of the switch."""
        return self._state

    async def async_turn_off(self, **kwargs: Any) -> None:
        """Turn off the switch."""
        try:
            if self._key.startswith('out'):
                await self.reader.output(self._key, on=False)
            elif self._key.startswith('pwm'):
                await self.reader.pwm_output(self._key, on=False)
            else:
                await self.reader.toggle_event(self._key, on=False)
            self._state = False
        except LK3Error:
            _LOGGER.error('An error occurred while turning off LK3 switch')
            self._available = False
        self.async_schedule_update_ha_state()

    async def async_turn_on(self, **kwargs: Any) -> None:
        """Turn on the switch."""
        try:
            if self._key.startswith('out'):
                await self.reader.output(self._key, on=True)
            elif self._key.startswith('pwm'):
                await self.reader.pwm_output(self._key, on=True)
            else:
                await self.reader.toggle_event(self._key, on=True)
            self._state = True
        except LK3Error:
            _LOGGER.error('An error occurred while turning on LK3 switch')
            self._available = False
        self.async_schedule_update_ha_state()

    async def _lk3_update(self) -> None:
        """Update LK3 entity."""
        self._state = self.reader.device.state[self._key]
